def fromstring(message):
  """Convert a Rejseplanen API error message into an appropriate exception.
  """
  error_code = message.split()[0]
  message = ERRORS.get(error_code, message)
  first_char = error_code[0]

  if first_char == 'R':
    return RestError(message)
  elif first_char == 'S':
    return ServerError(message)
  elif first_char == 'H':
    return TripServiceError(message)
  else:
    raise JourneyPlannerError(message)

#-------------------------------------------------------------------------------

ERRORS = {
  #General ReST Request Errors
  'R0001': 'Unknown service',
  'R0002': 'Invalid or missing request parameters',
  'R0007': 'Internal communication error',
  # Backend Server Errors
  'S1': 'The desired connection to the server could not be established or was not stable.',
  # Connection Request Errors
  'H9380': 'Dep./Arr./Intermed defined more than once',
  'H9360': 'Error in date field',
  'H9320': 'The input is incorrect or incomplete',
  'H9300': 'Unknown arrival station',
  'H9280': 'Unknown intermediate station',
  'H9260': 'Unknown departure station',
  'H9240': 'Unsuccessful search',
  'H9230': 'An internal error occurred',
  'H9220': 'Nearby to the given address stations could not be found',
  'H900': 'Unsuccessful or incomplete search (timetable change)',
  'H892': 'Inquiry too complex (try entering less intermediate stations)',
  'H891': 'No route found (try entering an intermediate station)',
  'H890': 'No connections found',
  'H895': 'Departure/Arrival are too near',
}

class JourneyPlannerError(Exception):
  """There was an exception that occurred while handling your request."""
  pass

class AuthenticationError(JourneyPlannerError):
    """Authentication failed."""
    pass

class RestError(JourneyPlannerError):
  """General rest error occured"""
  pass

class ServerError(JourneyPlannerError):
  """Backend Server gave error"""
  pass

class TripServiceError(JourneyPlannerError):
  """There was an error with your trip request"""
  pass



from . import JourneyPlanner

import argparse

parser = argparse.ArgumentParser(prog='python -m journeyplanner', description='Make requests to the journeyplanner API')

parser.add_argument('-u', help='authenticate request')
parser.add_argument('-p', help='authenticate request')

subparsers = parser.add_subparsers(help='api service', dest='service')

parser_location = subparsers.add_parser('location', help='location service')
parser_location.add_argument('query', type=str, help='the search query')

parser_trip = subparsers.add_parser('trip', help='trip service')
exorigin = parser_trip.add_mutually_exclusive_group(required=True)
exorigin.add_argument('--originid', type=int, help='origin')
exorigin.add_argument('--origin', nargs=2, type=float, metavar=('LATITUDE', 'LONGITUDE'))
exdest = parser_trip.add_mutually_exclusive_group(required=True)
exdest.add_argument('--destinationid', type=int, help='destination')
exdest.add_argument('--destination', nargs=2, type=float, metavar=('LATITUDE', 'LONGITUDE'))

parser_stopsnearby = subparsers.add_parser('stopsnearby', help='stops nearby service')
parser_stopsnearby.add_argument('latitude', type=float, help='latitude')
parser_stopsnearby.add_argument('longitude', type=float, help='longitude')

jp = JourneyPlanner()
args = parser.parse_args()
if args.u and args.p:
    jp.authenticate(args.u, args.p)

if args.service == 'location':
    for x in jp.location(args.query):
        print(str(x))

if args.service == 'stopsnearby':
    for x in jp.stopsnearby(latitude=args.latitude, longitude=args.longitude, radius=10000):
        print(str(x))

if args.service == 'trip':
    origin = args.originid if args.originid else args.origin
    destination = args.destinationid if args.destinationid else args.destination

    for x in jp.trip(origin=origin, destination=destination):
        print(str(x))



from .config import BASE_URL_OPEN, BASE_URL_AUTHENTICATED
from .error import AuthenticationError, JourneyPlannerError, fromstring as errorfromstring

from xml.etree import ElementTree


from urllib.request import build_opener, HTTPPasswordMgrWithDefaultRealm, HTTPBasicAuthHandler, HTTPError
from urllib.parse import urlencode

class Requester:
    def __init__(self):
        self.baseurl = BASE_URL_OPEN
        self.opener = build_opener()

    def authenticate(self, username, password):
        manager = HTTPPasswordMgrWithDefaultRealm()
        manager.add_password(None, BASE_URL_AUTHENTICATED, username, password)
        handler = HTTPBasicAuthHandler(manager)

        self.baseurl = BASE_URL_AUTHENTICATED
        self.opener.add_handler(handler)

    def get(self, service, **params):
        tuples = sorted(params.items(), key=lambda x: x[0])
        tuples = [t for t in tuples if t[1] is not None]
        querystring = urlencode(tuples)

        url = '{}/{}?{}'.format(self.baseurl, service, querystring)

        try:
            response = self.opener.open(url)

            return _parse(response)
        except HTTPError as error:
            if error.code == 401:
                raise AuthenticationError('Invalid credentials')


def _parse(file):
    try:
        tree = ElementTree.parse(file)
        root = tree.getroot()
        errormessage = root.get('error')

        if errormessage:
            raise errorfromstring(errormessage)

        return root
    except ElementTree.ParseError:
        raise JourneyPlannerError('Invalid server response') from None



from journeyplanner import JourneyPlanner, error, Endpoint
from journeyplanner.request import _parse

from io import StringIO

import unittest

class TestService(unittest.TestCase):

    def setUp(self):
        self.jp = JourneyPlanner()

    def test_fail_authentication(self):
        jp = JourneyPlanner()
        jp.authenticate('foo', 'bar')
        with self.assertRaises(error.AuthenticationError):
            next(jp.location('elmegade 5 københavn'))

    def test_location(self):
        first = next(self.jp.location('elmegade 5 københavn'))
        self.assertEqual(first.latitude, 55.689541)
        self.assertEqual(first.longitude, 12.558038)

    def test_trip(self):
        origin=Endpoint(latitude=55.68954, longitude=12.558038)
        first = next(self.jp.trip(origin=6, destination=8000, bus=False))
        self.assertEqual(first.valid, True)

    def test_stopsnearby(self):
        first = next(self.jp.stopsnearby(latitude=55.68954, longitude=12.558038))
        self.assertEqual(first.latitude, 55.689388)

if __name__ == '__main__':
    unittest.main()
